from setuptools import setup

setup(name='aiinpy',
      version='0.0.17',
      description='An artificial intelligence library?',
      license='MIT',
      packages=['aiinpy'],
      url='https://github.com/SeanMabli/aiinpy',
      install_requires=['numpy', 'alive_progress'])